var child_process = require('duplex-child-process')
var util = require('util')
var streamTo = require('stream-to')
var saveTo = require('save-to')
var join = require('path').join
var debug = require('debug')('gm')

util.inherits(gm, child_process)

module.exports = gm

function gm(source, options) {
  if (!(this instanceof gm))
    return new gm(source, options)

  this._initialize(source, options)
}

gm.extend = function (options) {
  function gm(source, options) {
    if (!(this instanceof gm))
      return new gm(source, options)

    this._initialize(source, options)
  }

  util.inherits(gm, this)
  gm.extend = this.extend

  if (typeof options === 'object') {
    Object.keys(options).forEach(function (key) {
      gm.prototype[key] = options[key]
    })
  }

  return gm
}

// default stdin source
gm.prototype.source = '-'

// default stdout output
gm.prototype.output = '-'

// default command
gm.prototype._command = 'convert'

// default path. prepended to the command.
gm.prototype.path = ''

// graphicsmagick by default
gm.prototype.imagemagick = !!process.env.GM_IMAGEMAGICK

gm.prototype._initialize = function (source, options) {
  if (typeof source === 'object' && source.constructor === Object) {
    options = source
    source = null
  } else if (!options) {
    options = {}
  }

  child_process.call(this, options)

  // setup options
  // options.source can only be string -
  // ie to overwrite `-`.
  if (typeof options.source === 'string')
    this.source = options.source

  // setup source
  if (source && source._readableState) {
    source.pipe(this._writer)
  } else if (Buffer.isBuffer(source)) {
    this._writer.end(source)
  } else if (Array.isArray(source)) {
    source.forEach(function (buf) {
      this._writer.write(buf)
    }, this)
    this._writer.end()
  } else if (typeof source === 'string') {
    this.source = source
  }

  debug('source: ' + this.source)

  return this
}

gm.prototype.command = function (command) {
  this._command = command
  return this
}

gm.prototype.operator = function () {
  var operators = this.operators = this.operators || []
  operators.push.apply(operators, arguments)
  return this
}

gm.prototype.setting = function () {
  var settings = this.settings = this.settings || []
  settings.push.apply(settings, arguments)
  return this
}

// setup the arguments for a convert
gm.prototype.convert = function (options) {
  var args = this.settings
  .concat(this.source)
  .concat(this.operators)
  .concat(this.output)

  debug('convert: ' + args.join(' '))

  this.spawn('convert', args, options)

  return this
}

// Overwrite child_process' spawn to allow buffering
var spawn = gm.prototype.spawn
gm.prototype.spawn = function (command, args, options) {
  // If for some reason there are no args
  if (!Array.isArray(args)) {
    options = args
    args = []
  }

  // GM support
  if (!this.imagemagick) {
    args.unshift(command)
    command = 'gm'
  }

  // Optional path
  if (this.path)
    command = join(this.path, command)

  // Setup the passthroughs if deleted
  if (!this._writer)
    child_process.call(this, options)

  debug('spawn: ' + command + ' ' + args.join(' '))

  spawn.call(this, command, args, options)

  var that = this

  // Setup the buffering option
  var buffer = this._buffer
  delete this._buffer
  if (buffer === true) {
    // Save as buffer in memory
    this.buffering = true
    streamTo.array(this._writer, function (err, arr) {
      if (err)
        return that.emit('error', err)

      that.input = arr
      that.buffered = true
      that.buffering = false
      that.emit('buffered')
    })
  } else if (typeof buffer === 'string') {
    // Save to file
    // May run into race conditions
    this.buffering = true
    saveTo(this._writer, buffer, function (err) {
      if (err)
        return that.emit('error', err)

      that.source = buffer
      that.buffered = true
      that.buffering = false
      that.emit('buffered')
    })
  }

  // Use the buffer.
  // You shouldn't be using a source at construction
  // and the buffered buffer.
  if (Array.isArray(this.input)) {
    this.input.forEach(function (buf) {
      this.write(buf)
    }, this)
    this.end()
  }

  return this
}

gm.prototype.buffer = function (filename) {
  if (!this.buffered)
    this._buffer = filename === 'string'
      ? filename
      : true

  return this
}

gm.prototype.timeout = function (ms) {
  this.timeoutID = setTimeout(this.destroy.bind(this), ms)
  this.once('close', this.clearTimeout)
  return this
}

gm.prototype.clearTimeout = function (id) {
  clearTimeout(id || this.timeoutID)
  return this
}